@@ -1,20 +1,18 @@
-'''
-Copyright (C) 2020  Davis Remmel
-Copyright 2021 Robert Schroll
-
-This program is free software: you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation, either version 3 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program.  If not, see <https://www.gnu.org/licenses/>.
-'''
+# Copyright (C) 2020  Davis Remmel
+# Copyright 2021 Robert Schroll
+#
+# This program is free software: you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
 import logging
 import tempfile
@@ -33,15 +31,35 @@ from .constants import PDFHEIGHT, PDFWIDTH, PTPERPX, SPOOL_MAX
 
 log = logging.getLogger(__name__)
 
-def render(source, *, progress_cb=lambda x: None):
-    # Exports the self as a PDF document to disk
+def render(source, *,
+           progress_cb=lambda x: None,
+           expand_pages=True,
+           template_alpha=0.3,
+           only_annotated=False):
+    """
+    Render a source document as a PDF file.
 
-    # progress_cb will be called with a progress percentage between 0 and
-    # 100.  This percentage calculation is split 50% for the rendering
-    # of the lines and 50% merging with the base PDF file.  This callback
-    # also provides an opportunity to abort the process. If the callback
-    # raises an error, this function will take steps to abort gracefullly
-    # and pass the error upwards.
+    source: The reMarkable document to be rendered.  This may be
+              - A filename or pathlib.Path to a zip file containing the
+                document, such as is provided by the Cloud API.
+              - A filename or pathlib.Path to a root-level file from the
+                document, such as might be copied off the device directly.
+              - An object implementing the Source API.  See rmrl.sources
+                for examples and further documentation.
+    progress_cb: A function which will be called with a progress percentage
+                 between 0 and 100.  The first 50% indicate rendering the
+                 annotations, and the second the merging of these into the
+                 base PDF file.  If this callback raises an error, this
+                 function will abort gracefully and propagate the error up
+                 the stack.
+    expand_pages: Boolean value (default True) indicating whether pages
+                  should be made larger, to reflect the view provided by
+                  the reMarkable device.
+    template_alpha: Opacity of the template backgrounds in notebooks.  0
+                    makes the templates invisible, 1 makes them fully dark.
+    only_annotated: Boolean value (default False) indicating whether only
+                    pages with annotations should be output.
+    """
 
     vector=True  # TODO: Different rendering styles
     source = sources.get_source(source)
@@ -50,9 +68,6 @@ def render(source, *, progress_cb=lambda x: None):
     # differently.
     uses_base_pdf = source.exists('{ID}.pdf')
 
-    # Document metadata should already be loaded (from device)
-    # ...
-
     # Generate page information
     # If a PDF file was uploaded, but never opened, there may not be
     # a .content file. So, just load a barebones one with a 'pages'
@@ -77,7 +92,7 @@ def render(source, *, progress_cb=lambda x: None):
         page = document.DocumentPage(source, pages[i], i)
         if source.exists(page.rmpath):
             changed_pages.append(i)
-        page.render_to_painter(pdf_canvas, vector)
+        page.render_to_painter(pdf_canvas, vector, template_alpha)
         annotations.append(page.get_grouped_annotations())
         progress_cb((i + 1) / len(pages) * 50)
     pdf_canvas.save()
@@ -131,7 +146,7 @@ def render(source, *, progress_cb=lambda x: None):
         # just add the annotations and forget about the rest,
         # which are page geometry transformations.
         if uses_base_pdf:
-            merge_pages(basepage, rmpage, i in changed_pages)
+            merge_pages(basepage, rmpage, i in changed_pages, expand_pages)
 
         progress_cb(((i + 1) / rmpdfr.numPages * 50) + 50)
 
@@ -146,9 +161,16 @@ def render(source, *, progress_cb=lambda x: None):
         else:
             basepdfr.Root.OCProperties = ocgprop
 
-    pdfw = PdfWriter()
     stream = tempfile.SpooledTemporaryFile(SPOOL_MAX)
-    pdfw.write(stream, basepdfr)
+    pdfw = PdfWriter(stream)
+    if not only_annotated:
+        # We are writing out everything, so we can take this shortcut:
+        pdfw.write(trailer=basepdfr)
+    else:
+        for i, page in enumerate(basepdfr.pages):
+            if i in changed_pages:
+                pdfw.addpage(page)
+        pdfw.write()
     stream.seek(0)
 
     log.info('exported pdf')
@@ -394,7 +416,7 @@ def apply_annotations(rmpage, page_annot, ocgorderinner):
             rmpage.Annots.append(pdf_a)
 
 
-def merge_pages(basepage, rmpage, changed_page):
+def merge_pages(basepage, rmpage, changed_page, expand_pages):
     # The general appraoch is to keep the base PDF. So, all
     # operations must be made upon the basepage. PyPDF2 will
     # keep all those pages' metadata and annotations,
@@ -411,21 +433,33 @@ def merge_pages(basepage, rmpage, changed_page):
     # ...
 
     # There is a bug here that can be seen with the NH file
+    # reMarkable uses the CropBox if it exists, otherwise
+    # the MediaBox.  (Though it does screw up the offsets
+    # for some combinations of rotations and landscape modes.)
     # It is possible (why?) for a page not to have a
     # MediaBox, so one must be taken from the parent. The
     # rM adds a bit to the width AND the height on this
     # file.
-    bpage_box = basepage.MediaBox
-    if not bpage_box:
-        # Should probably check if the parent has a mediabox
-        bpage_box = basepage.Parent.MediaBox
-    bpage_w = float(bpage_box[2]) - float(bpage_box[0])
-    bpage_h = float(bpage_box[3]) - float(bpage_box[1])
+    bpage_box = list(map(float, basepage.CropBox
+                                or basepage.MediaBox
+                                or basepage.Parent.MediaBox))
+
+    # Fix any malformed PDF that has a CropBox extending outside of
+    # the MediaBox, by limiting the area to the intersection.
+    if basepage.MediaBox:
+        for i, op in enumerate((max, max, min, min)):
+            bpage_box[i] = op(float(basepage.MediaBox[i]), bpage_box[i])
+
+    bpage_w = bpage_box[2] - bpage_box[0]
+    bpage_h = bpage_box[3] - bpage_box[1]
     # Round because floating point makes it prissy
     bpage_ratio = round(bpage_w / bpage_h * 10000) / 10000
     landscape_bpage = False
     if bpage_w > bpage_h:
         landscape_bpage = True
+        bpage_ratio = 1 / bpage_ratio # <= 1 always
+    if basepage.Rotate in ('90', '270'):
+        landscape_bpage = not landscape_bpage
 
     # If the base PDF page was really wide, the rM rotates
     # it -90deg (CCW) on the screen, but doesn't actually
@@ -434,14 +468,99 @@ def merge_pages(basepage, rmpage, changed_page):
     # the Web UI export. So, we must actually rotate the rM
     # page 90deg (CW) to fit on these wide pages.
 
-    rpage_box = rmpage.MediaBox
-    rpage_w = float(rpage_box[2]) - float(rpage_box[0])
-    rpage_h = float(rpage_box[3]) - float(rpage_box[1])
+    # Since we create this page, we know there isn't a different
+    # CropBox to worry about.  We also know width < height
+    rpage_box = list(map(float, rmpage.MediaBox))
+    rpage_w = rpage_box[2] - rpage_box[0]
+    rpage_h = rpage_box[3] - rpage_box[1]
     rpage_ratio = rpage_w / rpage_h
-    if landscape_bpage:
-        rmpage.Rotate = 90
-        rpage_ratio = rpage_h / rpage_w
 
+    effective_rotation = int(basepage.Rotate)
+    # If the page is landscape, reMarkable adds a -90 degree rotation.
+    if landscape_bpage:
+        effective_rotation = (effective_rotation + 270) % 360
+    # The rmpage picks up the rotation of the base page -- that is,
+    # its own rotation is relative to the basepage.  We don't want
+    # any net rotation, so we rotate it backwards now, so that with
+    # the basepage rotation, it ends up upright.
+    rmpage.Rotate = (360 - effective_rotation) % 360
+
+    if effective_rotation in (0, 180):
+        flip_base_dims = False
+    elif effective_rotation in (90, 270):
+        flip_base_dims = True
+    else:
+        assert False, f"Unexpected rotation: {effective_rotation}"
+
+    if bpage_ratio <= rpage_ratio:
+        # These ratios < 1, so this indicates the basepage is more
+        # narrow, and thus we need to extend the width.  Extra space
+        # is added to the right of the screen, but that ends up being
+        # a different page edge, depending on rotation.
+        if not flip_base_dims:
+            new_width = rpage_ratio * bpage_h
+            scale = bpage_h / rpage_h
+            if effective_rotation == 0:
+                bpage_box[2] = new_width + bpage_box[0]
+            else:
+                bpage_box[0] = bpage_box[2] - new_width
+        else:
+            # Height and width are flipped for the basepage
+            new_height = rpage_ratio * bpage_w
+            scale = bpage_w / rpage_h
+            if effective_rotation == 90:
+                bpage_box[3] = new_height + bpage_box[1]
+            else:
+                bpage_box[1] = bpage_box[3] - new_height
+    else:
+        # Basepage is wider, so need to expand the height.
+        # Extra space is added at the bottom of the screen.
+        if not flip_base_dims:
+            new_height = 1/rpage_ratio * bpage_w
+            scale = bpage_w / rpage_w
+            if effective_rotation == 0:
+                bpage_box[1] = bpage_box[3] - new_height
+            else:
+                bpage_box[3] = new_height + bpage_box[1]
+        else:
+            # Height and width are flipped for the basepage
+            new_width = 1/rpage_ratio * bpage_h
+            scale = bpage_h / rpage_w
+            if effective_rotation == 90:
+                bpage_box[2] = new_width + bpage_box[0]
+            else:
+                bpage_box[0] = bpage_box[2] - new_width
+
+    if expand_pages:
+        # Create a CropBox, whether or not there was one before.
+        basepage.CropBox = bpage_box
+        if not basepage.MediaBox:
+            # Provide a MediaBox, in the odd case where there isn't one.
+            basepage.MediaBox = bpage_box
+        else:
+            # Expand the MediaBox as necessary to include the entire CropBox.
+            for i, op in enumerate((min, min, max, max)):
+                basepage.MediaBox[i] = op(float(basepage.MediaBox[i]), bpage_box[i])
+
+    # If this wasn't a changed page, don't bother with the
+    # following.
+    if not changed_page:
+        return
+
+    # Scale and (if necesssary) rotate the notebook page
+    # and overlay it to the basepage. Might have to push
+    # it a bit, depending on the direction.
+    np = PageMerge(basepage).add(rmpage)
+
+    # Move the overlay page to be based on the coordinates
+    # of the base page CropBox
+    np[1].x = bpage_box[0]
+    np[1].y = bpage_box[1]
+    np[1].scale(scale)
+
+    #TODO: Test all of these annotations with various rotations
+    # and offsets.
+    if landscape_bpage:
         # Annotations must be rotated because this rotation
         # statement won't hit until the page merge, and
         # pdfrw is unaware of annotations.
@@ -454,63 +573,8 @@ def merge_pages(basepage, rmpage, changed_page):
                     rect[3],
                     PDFWIDTH - rect[2]])
 
-
-    # Resize the base page to the notebook page ratio by
-    # adjusting the trimBox. If the basepage was landscape,
-    # the trimbox must expand laterally, because the rM
-    # rotates the page on-screen into portrait. If the
-    # basepage was already portrait, it must expand
-    # laterally.
-
-    adjust = 0
-    if bpage_ratio <= rpage_ratio:
-        # Basepage is taller, so need to expand the width.
-        # The basepage should be pushed to the right, which
-        # is also the top of the rM in portrait mode. A
-        # push to the right is really just decreasing the
-        # left side.
-        new_width = rpage_ratio * bpage_h
-        if landscape_bpage:
-            adjust = float(bpage_box[2]) - new_width
-            bpage_box[0] = adjust
-        else:
-            # Portrait documents get pushed to the left, so
-            # expand the right side.
-            adjust = float(bpage_box[0])
-            bpage_box[2] = new_width + float(bpage_box[0])
-    elif bpage_ratio > rpage_ratio:
-        # Basepage is fatter, so need to expand the height.
-        # The basepage should be pushed to the top, which is
-        # also the top of the rM in portrait mode. A push to
-        # the top is really decreasing the bottom side.
-        new_height = (1 / rpage_ratio) * bpage_w
-        adjust = float(bpage_box[3]) - new_height
-        bpage_box[1] = adjust
-
-    # If this wasn't a changed page, don't bother with the
-    # following.
-    if not changed_page:
-        return
-
-    # Scale and (if necesssary) rotate the notebook page
-    # and overlay it to the basepage. Might have to push
-    # it a bit, depending on the direction.
-    #basepage.Rotate = -90
-    np = PageMerge(basepage).add(rmpage)
-
     annot_adjust = [0, 0]
 
-    if bpage_ratio <= rpage_ratio:
-        scale = bpage_h / np[1].h
-        np[1].scale(scale)
-        np[1].x = adjust
-        annot_adjust[0] = adjust
-    elif bpage_ratio > rpage_ratio:
-        scale = bpage_w / np[1].w
-        np[1].scale(scale)
-        np[1].y = adjust
-        annot_adjust[1] = adjust
-
     if '/Annots' in rmpage:
         for a, annot in enumerate(rmpage.Annots):
             rect = annot.Rect
